/*
 * [Span.java]
 *
 * Summary: Determine which chars are used in a string.
 *
 * Copyright: (c) 2002-2017 Roedy Green, Canadian Mind Products, http://mindprod.com
 *
 * Licence: This software may be copied and used freely for any purpose but military.
 *          http://mindprod.com/contact/nonmil.html
 *
 * Requires: JDK 1.8+
 *
 * Created with: JetBrains IntelliJ IDEA IDE http://www.jetbrains.com/idea/
 *
 * Version History:
 *  1.0 2011-01-03 initial version
 */
package com.mindprod.quoter;

import com.mindprod.common18.ST;
import com.mindprod.fastcat.FastCat;

import java.util.BitSet;
import java.util.regex.Pattern;

/**
 * Determine which chars are used in a string.
 * <p/>
 * Useful in composing regexes to hop over text.
 *
 * @author Roedy Green, Canadian Mind Products
 * @version 4.5 2009-02-26 add both Java string quoting and plain for Java search/regexes.
 * @since 2002-06-19
 */
final class Span extends TextProcessor
    {
    /**
     * lowest char we track.
     */
    private static final int LOWER_LIMIT = 32;

    /**
     * highest char we track +1
     */
    private static final int UPPER_LIMIT = 128;

    Pattern p = Pattern.compile(
            "space !\"#\\$%&'\\(\\)\\*\\+,\\-\\./0123456789:;" +
            "<=>\\?@ABCDEFGHIJKLMNOPQRSTUVWXYZ\\[\\\\\\]\\^_`abcdefghijklmnopqrstuvwxyz\\{\\|\\}~"
    );

    /**
     * display char as it would appear in a search regex, not in Java string
     *
     * @param c the char to display
     *
     * @return how this char displayed in a regex
     */
    private static String displayChar( char c )
        {
        switch ( c )
            {
            case '$':
                return "\\$";
            case '&':
                return "\\&"; // intelliJ says not reserved but && is intersection.
            case '(':
                return "\\(";
            case ')':
                return "\\)";
            case '*':
                return "\\*";
            case '+':
                return "\\+";
            case '-':
                return "\\-";
            case '.':
                return "\\.";
            case '?':
                return "\\?";
            case '[':
                return "\\[";
            case '\\':
                return "\\\\";
            case ']':
                return "\\]";
            case '^':
                return "\\^";   // means "not" otherwise
            case '|':
                return "\\|";
            case '{':
                return "\\{";
            case '}':
                return "\\}";
            default:
                if ( c < ' ' || c == 0x7f || c == 0xff )
                    {
                    return "\\x" + ST.toLZHexString( c, 2 );
                    }
                else
                    {
                    return String.valueOf( c );
                    }
            }
        }

    /**
     * display to chars marked on, compacting into ranges if possible
     *
     * @param used bit set of chars to include
     *
     * @return list of char marked on, possibly compacted into ranges.
     */
    private String displayRanges( final BitSet used )
        {
        StringBuilder sb = new StringBuilder( UPPER_LIMIT );
        // display chars used
        for ( int i = LOWER_LIMIT; i < UPPER_LIMIT; i++ )
            {
            if ( used.get( i - LOWER_LIMIT ) )
                {
                // lookahead for last used in this run, then we can display it as \\p{Lower}
                int lastUsed = 4095; // in case run off the end.
                inner:
                for ( int j = i + 1; j < 4096; j++ )
                    {
                    if ( !used.get( j - LOWER_LIMIT ) )
                        {
                        lastUsed = j - 1; // might be i
                        break inner;
                        }
                    } // end inner for
                if ( lastUsed > i + 2 )  // leave abc but collapse abcd to a-d
                    {
                    // can handle with a range.
                    sb.append( displayChar( ( char ) i ) );
                    sb.append( '-' );
                    sb.append( displayChar( ( char ) lastUsed ) );
                    i = lastUsed;
                    }
                else
                    {
                    sb.append( displayChar( ( char ) i ) );
                    }
                } // end if used
            } // end outer for
        return sb.toString();
        }

    /**
     * Converts text, will be overridden. can handle null and empty strings.
     *
     * @param raw text to be translated.
     *
     * @return String representing the cooked output, possibly null.
     */
    String process( String raw )
        {
        assert raw != null : "TextProcessor.process raw must not be null";
        // determine which chars used
        final BitSet used = new BitSet( UPPER_LIMIT - LOWER_LIMIT );
        boolean including = true;
        raw = ST.trimLeading( raw, "\n\r" );  // leave spaces in place
        raw = ST.trimTrailing( raw, "\n\r" );
        for ( int i = 0; i < raw.length(); i++ )
            {
            final char c = raw.charAt( i );
            if ( c == '\n' )
                {
                /** now into the second row of exclusions */
                including = false;
                }
            else
                {
                if ( LOWER_LIMIT <= c && c < UPPER_LIMIT )
                    {
                    used.set( c - LOWER_LIMIT, including );
                    }
                }
            }
        final FastCat sb = new FastCat( 7 );
        sb.append( "Only characters in the range space to ~ are shown.\n" );
        sb.append( "All characters in the input are treated literally. e.g. - is just a dash, " +
                   "not a range specifier.\n" );
        sb.append( "Including the first row, then excluding the second:\n  [" );
        sb.append( displayRanges( used ) );
        sb.append( "]\nalternatively:\n  [^" );
        used.flip( 0, UPPER_LIMIT - LOWER_LIMIT - 1 );
        sb.append( displayRanges( used ) );
        sb.append( "]\n" );
        return sb.toString();
        }
    }
